//
// chipset_mgr.c - the feature / chipset class definitions
//   by Ed Porras <ep@bangchemistry.com>
//   with portions of code from
//
//     wmgtemp by Roger Dunce <kronos@nanunanu.org> 
//
//
// $Log: chipset_mgr.c,v $
// Revision 1.12  2003/03/25 02:14:24  ep
// moved cs_feature and chipset into their own files.. updated code
// to conform to gcc3 std
//
// Revision 1.11  2001/11/13 06:36:20  ep
// redeclared destructors as public.. minor sanity assignment
// for unkown f_type in switch..
//
// Revision 1.10  2001/11/12 08:10:54  ep
// reimplemented purge by using STL algorithm.. made
// destructor public..
//
// Revision 1.9  2001/11/12 02:53:26  ep
// added more sanity checks for updating mapped feature values..
// renamed checkLast to purgeLast.. made output more readable..
//
// Revision 1.8  2001/11/09 07:39:01  ep
// added api call to check if features should be ignored..
// removed ALARM type.. modified code so only features that have
// others mapped to it are maintained.. fixed stupid issue
// with xor vs 2's comp.. added warning for features that are
// mapped more than once
//
// Revision 1.7  2001/11/08 16:56:19  ep
// modified verbose messages to get better mapping info
//
// Revision 1.6  2001/11/08 08:09:07  ep
// removed hoaky extension methods.. shouldn't be handled here
//
// Revision 1.5  2001/11/07 08:42:08  ep
// added verbose message during feature detect.. fixed bug
// for temp detect based on bits set
//
// Revision 1.4  2001/11/06 06:58:46  ep
// fixed initializer list
//
// Revision 1.3  2001/11/06 06:52:29  ep
// the beginnings of the better detect mechanism
//
// Revision 1.2  2001/11/05 08:04:16  ep
// implemented data retrieval method (uses single index)..
// added poll bit modes
//
// Revision 1.1  2001/11/05 05:24:26  ep
// Initial revision
//
//

#include <iostream>
#include <iomanip>
#include <string>
#include <cstring>
#include <vector>
#include <algorithm>
#include <sensors/sensors.h>
#include "chipset_mgr.h"

using namespace std;

namespace wmmsens
{
  // static bool initialization
  bool chipset_mgr::verbose = false;

  // =========================================================================
  // the chipset manager class

  // 
  chipset_mgr::~chipset_mgr()
  {
    for (register ui32 i = 0; i < chipset_list.size(); i++)
      delete chipset_list[i];
  }

  //
  // detect all available chipsets
  bool chipset_mgr::detect(ui16 poll_bits)
  {
    int chip_nr = 0;
    bool added = false;

    // one by one, fetch all chipsets 
    do
    {
      const sensors_chip_name* name = sensors_get_detected_chips(&chip_nr);

      // NULL means we're done
      if (name == NULL)
	break;

      // try to detect all the chip features
      if (registerChipset( *name, poll_bits ))
	added = true;
    }
    while (true);

    // output the ones we kept
    if (added && verbose)
    {
      cout << *this << endl;

      cout << " +- if you see more chipsets than expected, lm_sensors -+" 
	   << endl
	   << " +- may be detecting the same one more than once.      -+" 
	   << endl
	   << " +-              check your modprobes                  -+" 
	   << endl;
    }
    return added; 
  }


  //
  // this is called for each chipset to get its features
  bool chipset_mgr::registerChipset(const sensors_chip_name& chip, 
				    ui16 poll_bits)
  {
    // create an instance for the chipset
    chipset* cs = new chipset( chip, poll_bits );

    // output its name & info 
    if (verbose)
      cout << "+- Found \"" << chip.prefix 
	   << "\" chipset on \"" 
	   << sensors_get_adapter_name(chip.bus) << "\" -+" << endl;

    // and get its features.. if we get any that we expect, we save it
    if (cs->getFeatures( history_steps ))
    {
      // store it
      chipset_list.push_back( cs );

      // and force update of all features to get initial readings
      cs->update(true);

      // increment the total count of features
      total_features += cs->numFeatures();

      // just to separate output of chipsets
      if (verbose)
	cout << endl;
      return true;
    }

    // otherwise, discard it
    if (verbose)
      cout << " discarding chipset (does not seem relevant)" << endl << endl;

    delete cs;
    return false;
  }


  //
  // returns the cs_feature in the chipset / pos offset computed from
  // idx
  const cs_feature& chipset_mgr::operator[](ui32 idx) const
  {
    idx %= total_features;
    ui32 cs_cnt = chipset_list.size();

    for (register ui32 i = 0; i < chipset_list.size(); i++, cs_cnt--)
      if (idx >= chipset_list[i]->numFeatures())
	idx -= chipset_list[i]->numFeatures();

    return (*chipset_list[cs_cnt])[idx];
  }

  //
  // output op
  ostream& operator<<(ostream &o, const chipset_mgr& cm)
  {
    for (register ui32 i = 0; i < cm.chipset_list.size(); i++)
      o << *cm.chipset_list[i] << endl;
    return o;
  }

} // namespace wmmsens
