//
// cs_feature.c - the feature / chipset class definitions
//   by Ed Porras <ep@bangchemistry.com>
//   with portions of code from
//
//     wmgtemp by Roger Dunce <kronos@nanunanu.org> 
//
//
// $Log: cs_feature.c,v $
// Revision 1.1  2003/03/25 02:14:20  ep
// Initial revision
//
//

#include <iostream>
#include <iomanip>
#include <string>
#include <cstring>
#include <vector>
#include <algorithm>
#include <sensors/sensors.h>
#include "cs_feature.h"

using namespace std;

namespace wmmsens
{
  // =========================================================================
  // the chipset feature class

  //
  // constructor
  cs_feature::cs_feature(const sensors_chip_name& name, 
			 const sensors_feature_data& fd,
			 csf_type t, ui8 steps) :
    chipname(name), f_data(fd), 
    val( steps ), min_read(-1), max_read(-1), 
    f_type(t),
    min_idx(-1), max_idx(-1), alarm_on( false )
  {
    // copy the feature name (need to allocate storage)
    f_data.name = new char[ strlen( fd.name ) + 1 ];
    strcpy( (char *) f_data.name, fd.name );

    // get the feature's label to use for display
    char *lbl_str;
    if (sensors_get_label(chipname, f_data.number, &lbl_str) == 0)
    {
      lbl = lbl_str;
      free(lbl_str);
    }  
  }


  //
  // destructor
  cs_feature::~cs_feature() 
  { 
    // free all mapped features
    for (register ui32 i = 0; i < mapped_feature.size(); i++)
      delete mapped_feature[i];

    // and the name & history
    delete [] f_data.name;
  }


  //
  // polls the hardware for the latest value
  void cs_feature::update(bool all)
  {
    if (all)
    {
      // update the values of the mapped features as well.. this
      // should only be done once..
      for (register ui32 i = 0; i < mapped_feature.size(); i++)
	mapped_feature[i]->update(false);
    }

    // update the feature by polling lm_sens
    double cur;
    sensors_get_feature(chipname, id(), &cur);

    // if we're maintaining a history, handle the rotation and compute
    // the min & max
    if (history())
    {
      featval high = 0, low = 10000;

      // shift the array and compute min & max while we're iterating
      // through it
      for (register ui32 i = 0; i < val.size() - 1; i++)
      {
	val[i] = val[i + 1];

	if (val[i] > high)
	  high = val[i];

	if (val[i] < low && val[i] > 0)
	  low = val[i];
      }

      // count that one as part of the min & max..
      if (cur > high)
	high = cur;

      if (cur < low && cur > 0)
	low = cur;

      // update the run high / low values
      max_read = high;
      min_read = low;
    }

    // now store the value just read
    val[ val.size() - 1 ] = static_cast<featval>(cur);

    // now check the value and mark an alarm if out of range
    // !! fix me !! need to add checks to ensure min_idx and max_idx
    // have been set
    if (mapped_feature.size() > 0)
      switch (f_type)
      {
      case VOLTAGE:
	if (min_idx == -1 || max_idx == -1) // paranoia
	  break;
	alarm_on = (value() <= mapped_feature[min_idx]->value() ||
		    value() >= mapped_feature[max_idx]->value());
	break;

      case FAN:
	if (min_idx == -1) // paranoia
	  break;
	alarm_on = ( value() <= mapped_feature[min_idx]->value() );
	break;

      case TEMPERATURE:
	if (min_idx == -1 || max_idx == -1) // paranoia
	  break;

	// according to the docs, when temperatures exceed the
	// "over" setting, the alarm is triggered.. this is not
	// reset until the value is below the "hyst" setting
	if (alarm_on)
	{
	  if (value() <= mapped_feature[min_idx]->value())
	    alarm_on = false;
	}
	else
	  if (value() >= mapped_feature[max_idx]->value())
	    alarm_on = true;
	break;
      }
  }

  //
  // adds a feature to the mapped list
  bool cs_feature::addMappedFeature(cs_feature& mf)
  {
    bool warn = false;
    bool keep = true;

    // based on certain keywords, well try to guess the limits
    std::string f_name = mf.name();

    switch (f_type)
    {
    case VOLTAGE:
      if (f_name.find("min") != string::npos)
      {
	warn = (min_idx != -1);

	min_idx = mapped_feature.size();
      }
      else if (f_name.find("max") != string::npos)
      {
	warn = (max_idx != -1);

	max_idx = mapped_feature.size();
      }
      else 
	keep = false;
      break;

    case FAN:
      if (f_name.find("min") != string::npos)
      {
	warn = (min_idx != -1);

	min_idx = mapped_feature.size();
      }
      else
	keep = false;
      break;

    case TEMPERATURE:
      if (f_name.find("hyst") != string::npos)
      {
	warn = (min_idx != -1);

	min_idx = mapped_feature.size();
      }
      else if (f_name.find("over") != string::npos)
      {
	warn = (max_idx != -1);
	max_idx = mapped_feature.size();
      }
      else
	keep = false;
      break;

      // don't keep these...
    case UNKNOWN:
    default:
      keep = false;
      break;
    }

    if (keep)
    {
      if (warn)
      {
	cout << "++ WARNING: a feature similar to \"" 
	     << f_name << "\" has already been mapped to "
	     << "\"" << name() << "\"" << endl
	     << "++ This makes me feel dirty" << endl;
      }

      // store the feature in the mapped list
      mapped_feature.push_back(&mf);
      return true;
    }
    return false;
  }


  //
  // output op
  ostream& operator<<(ostream &o, const cs_feature& f)
  {
    o << "fid: " << f.id() 
      << ", name: \"" << f.name() << "\""
      << ", label: \"" << f.label() << "\""
      << ", init val: " << f.value() 
      << ", map: " << f.mapping() << endl;

    if (f.mapped_feature.size() > 0)
    {
      o << "    " << f.mapped_feature.size() << " mapped features: " << endl;
      for (register ui32 i = 0; i < f.mapped_feature.size(); i++)
	o << "    " << *f.mapped_feature[i];
    }
    return o;
  }

} // namespace wmmsens
