//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// $Log: sensor_mgr.c,v $
// Revision 1.12  2001/11/05 03:58:18  ep
// quick improvements before ver2.0
//
// Revision 1.11  2001/11/01 23:19:12  ep
// sensormgr now using vector of features.. renamed
// feature class
//
// Revision 1.10  2001/10/30 23:31:46  ep
// fixed dumb bug where min / max weren't being updated
//
// Revision 1.9  2001/10/19 01:51:34  ep
// minor mods dealing with constant names & comments
//
// Revision 1.8  2001/10/18 21:03:26  ep
// added detect routine.. made register method private
// and added typedef for double history (enabled with define)
//
// Revision 1.7  2001/10/18 19:09:54  ep
// feature object now copies chipset prefix also..
//
// Revision 1.6  2001/10/18 18:49:08  ep
// moved limit-detecting code to a private method.. fixed
// minor issues in the routine
//
// Revision 1.5  2001/10/18 17:41:01  ep
// added limit-detection (a feature-detecting feature, if you will)..
//
// Revision 1.4  2001/10/18 03:05:22  ep
// fixed file close in initsensors (in case file is not found)..
// minor changes to output
//
// Revision 1.3  2001/10/17 17:48:12  ep
// implemented feature-detect which looks for temp and fan
// entries.. renamed some constants and added some needed getters
//
// Revision 1.2  2001/10/17 15:02:34  ep
// fixed init of history.. added getters for min, man, and cur
// temp.. started code for feature detect
//
// Revision 1.1  2001/10/16 15:43:41  ep
// Initial revision
//
//

#include <iostream>
#include <stdexcept>
#include <vector>
#include <sensors/chips.h>
#include <sensors/sensors.h>
#include "sensor_mgr.h"
using namespace std;

// ==============================================================
// feature

//
// constructor / destructor
feature::feature(const sensors_chip_name& n, ui32 feat, double lim, 
		 ui32 steps) throw(length_error) :
    chipname(n), max_t(0), min_t(1000), limit_val(lim), history(NULL), 
    NUM_STEPS(steps), last_temp(NUM_STEPS - 1), feature_id(feat)
{
   // since we copied the chipset struct, we must doa deep copy to get
   // the name too (in case the struct gets replaced by another while
   // the mgr performs a search)
   chipname.prefix = new char[ strlen(n.prefix) + 1 ];
   strcpy( chipname.prefix, n.prefix );

   // create an array that big and set them all to 0
   history = new tempval[ NUM_STEPS ](0.0); 

#if 1
   // shouldn't the above (0.0) do this? For some reason it is not
   // working.. am I confused?
   for (register ui32 i = 0; i < NUM_STEPS; i++)
      history[i] = 0.0;
#endif

   // get the feature's label to use for display
   char *lbl_str;
   if (sensors_get_label(chipname, feature_id, &lbl_str) == 0)
   {
      lbl = lbl_str;
      free(lbl_str);
   }  

   // throw an exception if someone is being funny and passes 0 for
   // number of steps to store.. should this be at the end?
   if (steps == 0)
      throw length_error("Invalid num_steps: 0");
}


//
// update sensor data
void feature::update()
{
   // get the latest reading
   double cur;
   sensors_get_feature(chipname, feature_id, &cur);

   if (isHistory())
   {
      tempval high = 0, low = 5000;

      // shift the array and compute min & max while we're iterating
      // through it
      for (register ui32 i = 0; i < last_temp; i++)
      {
	 history[i] = history[i + 1];

	 if (history[i] > high)
	    high = history[i];

	 if (history[i] < low && history[i] > 0)
	    low = history[i];
      }

      // count that one as part of the min & max..
      if (cur > high)
	 high = cur;

      if (cur < low && cur > 0)
	 low = cur;

      // update the run high / low values
      max_t = high;
      min_t = low;
   }

   // store it..
   history[ last_temp ] = static_cast<tempval>(cur);
}


// ==============================================================
// sensor_mgr

//
// destructor.. 
sensor_mgr::~sensor_mgr()
{
   // delete all registered sensors
   for (register ui32 i = 0; i < sensor.size(); i++)
      delete sensor[i];
}


//
// register the chipset to keep track of.. this tries to determine the
// features and limits from it
bool sensor_mgr::registerChipset(const sensors_chip_name& chipset)
{
   enum { NUM_KEYS = 2 };
   static const string KEY_STR[] = { "temp", "fan" };
   static const bool   HIST_B[]  = {   true, false };  
   static const string LIM_KEY_STR[] = { "_over", "_min" };

   bool added = false;
   int nr1 = 0, nr2 = 0;

   // output the name of the sensor 
   cout << "Attempting to scan features of " << chipset.prefix 
	<< " on " << sensors_get_adapter_name(chipset.bus) << endl;

   // one by one..
   while (true)
   {
      const sensors_feature_data* sfd = sensors_get_all_features(chipset, 
								 &nr1, 
								 &nr2);
      // didn't find another one
      if (sfd == NULL)
	 break;

      // first, see if it's temperature- or fan-related
      string f = sfd->name;

      cout << "got feature \"" << f 
	   << "\" (" << sfd->number << ", map: " << 
	  sfd->mapping << ")" << endl;

      for (int i = 0; i < NUM_KEYS; i++)
      {
	 if ( f.find( KEY_STR[i] ) != string::npos )
	 {
	    // got something with the word "temp".. see if its
	    // something like temp1, temp2, etc.. or the same for
	    // "fan"
	    if (f.length() == KEY_STR[i].length() + 1)
	    {
	       // looking good.. now look for limits
	       double limit = getLimit(chipset, 
				       KEY_STR[i], LIM_KEY_STR[i], 
				       nr1, nr2);

	       // create a new instance and add it to the bunch
	       feature *f = new feature( chipset, 
					 sfd->number,
					 limit, 
					 (HIST_B[i] ? 
					  NUM_STEPS : 1) );

	       sensor.push_back( f );

	       cout << "  registered with label \"" 
		    << f->label() << "\"" << endl << endl;
	       added = true;
	    }
	 }
      }
   }
   return added;
}


//
// tries to determine the feature's limit(s).. must be passed an
// already-searching pair of nr values (see register method).. returns
// 0 if no limit is found
tempval sensor_mgr::getLimit(const sensors_chip_name& chipset, 
			     const string& key, const string& lim_str_suf, 
			     int &nr1, int &nr2)
{
   // keep scannig the features until we find one that "looks"
   // right.. we assume that limits follow the feature discovered
   // earlier.. this might be a bad asumption
   while (true)
   {
      const sensors_feature_data* sfd_l = sensors_get_all_features(chipset, 
								   &nr1, 
								   &nr2);

      cout << "got feature \"" << sfd_l->name
	   << "\" (" << sfd_l->number << ", map: " << 
	  sfd_l->mapping << ")" << endl;

      // check if we got something
      if (sfd_l == NULL)
	 break; // nope

      // get the name and see if we find the same
      // key as before
      string fn = sfd_l->name;

      int f_idx = fn.find(key);
      if (f_idx == -1)
	 break;

      // ok.. if it's a temp, look to see if it says
      // "over".. if it's a fan, look for "min"
      if ( fn.find(lim_str_suf, f_idx) != string::npos)
      {
	 double limit;

	 // found it.. get the value
	 if (sensors_get_feature(chipset, sfd_l->number, &limit) == 0)
	 {
	    cout << "  guessed limit key: \"" << fn 
		 << "\" and got: " << limit << endl;
	    return limit;
	 }
      }
   }
   // we failed..
   cout << "couldn't determine limit! assuming 0" << endl;
   return 0;
}

//
// try to detect chipsets
bool sensor_mgr::detect()
{
   int chip_nr = 0;
   const sensors_chip_name *name;
   bool added = false;

   do
   {
      name = sensors_get_detected_chips(&chip_nr);

      if (name == NULL)
	 break;

      if (registerChipset( *name ))
	 added = true;
      else
	 cout << " chipset \"" << name->prefix 
	      << "\" detected but ignored" << endl << endl;
   }
   while (true);

   cout << " +- if you see more chipsets than expected, lm_sensors -+" << endl
	<< " +- may be detecting the same one more than once.      -+" << endl
	<< " +-              check your modprobes                  -+" << endl;

   return added;
}
