#ifndef SENSOR_MGR_H
#define SENSOR_MGR_H

//
// two classes to manage sensor history
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// $Log: sensor_mgr.h,v $
// Revision 1.11  2001/11/05 03:58:18  ep
// quick improvements before ver2.0
//
// Revision 1.10  2001/11/01 23:19:12  ep
// sensormgr now using vector of features.. renamed
// feature class
//
// Revision 1.9  2001/10/30 23:31:46  ep
// fixed dumb bug where min / max weren't being updated
//
// Revision 1.8  2001/10/18 21:03:26  ep
// added detect routine.. made register method private
// and added typedef for double history (enabled with define)
//
// Revision 1.7  2001/10/18 19:09:54  ep
// feature object now copies chipset prefix also..
//
// Revision 1.6  2001/10/18 18:49:08  ep
// moved limit-detecting code to a private method.. fixed
// minor issues in the routine
//
// Revision 1.5  2001/10/18 17:39:19  ep
// renamed num_sens to num_feat for consistency.. added limit
// identifiers and modified constructor to accept argument for
// it
//
// Revision 1.4  2001/10/18 03:05:22  ep
// fixed file close in initsensors (in case file is not found)..
// minor changes to output
//
// Revision 1.3  2001/10/17 17:48:12  ep
// implemented feature-detect which looks for temp and fan
// entries.. renamed some constants and added some needed getters
//
// Revision 1.2  2001/10/17 15:02:34  ep
// fixed init of history.. added getters for min, man, and cur
// temp.. started code for feature detect
//
// Revision 1.1  2001/10/16 15:43:41  ep
// Initial revision
//
//

#include <string>
#include <stdexcept>
#include <vector>
#include <sensors/sensors.h>

//
// handy typedefs
typedef unsigned char  ui8;
typedef unsigned short ui16;
typedef unsigned int   ui32;

// define this if you want to maintain doubles for history values
#ifdef DBL_HIST_VAL
typedef double tempval;
#else
typedef float tempval;
#endif

//
// keeps track of history data for one sensor feature
class feature 
{
private:
   // the chipset struct.. this is a copy of the one returned by the
   // sensor lib..
   sensors_chip_name chipname;

   // history min & max.. only pertains to features that we keep
   // history of (temp)
   tempval max_t, min_t;

   // limit that triggers an alarm if crossed.. obtained from lmsensor
   // lib
   double limit_val;


   // if this is a history feature, we maintain an array of
   // n.. otherwise, it'll be of 1
   tempval* history;

   // number of steps..
   const ui32 NUM_STEPS,
       last_temp; // if history, this points to the last pos in the
                  // array.. otherwise, it points to 1 (at index 0)

   // the feature we're looking at
   const int feature_id;
   string lbl;

   // prohibit
   feature(const feature&);
   feature& operator=(const feature&);

public:
   // constructor
   feature(const sensors_chip_name& name, ui32 feat, double limit, 
	   ui32 steps) throw(length_error);
   ~feature() { delete [] history; delete [] chipname.prefix; }

   // getters
   tempval minVal() const { return min_t; }
   tempval maxVal() const { return max_t; }
   tempval curVal() const { return history[ last_temp ]; }
   tempval limit() const { return limit_val; }
   const string& label() const { return lbl; }
   bool isHistory() const { return (NUM_STEPS > 1); }

   // returns the temperature at the given index
   tempval operator[](ui32 idx) const throw(out_of_range) {
      if (idx >= NUM_STEPS)
	 throw out_of_range("Step out of range");
      return history[idx];
   }

   // reads latest data from sensor lib
   void update();
};



//
// maintains n number of features
class sensor_mgr
{
public:
   // number of stes for history-types
   const ui32 NUM_STEPS;

   // the features we're maintaining..
   vector<feature *> sensor;

   // prohibit
   sensor_mgr(const sensor_mgr&);
   sensor_mgr& operator=(const sensor_mgr&);

public:
   // constructor & destructor
   sensor_mgr(ui32 step_cnt) : 
       NUM_STEPS(step_cnt) { }
   ~sensor_mgr();

   // public methods
   ui32 numFeatures() const { return sensor.size(); }

   // tries to detect chipsets with features that have standard names
   bool detect();

   // updates the readings of all registered sensors
   void update() {
      for (register ui32 i = 0; i < numFeatures(); i++)
	 sensor[ i ]->update();
   }

   // returns the feature at the given index
   const feature& operator[](ui32 sn) const throw(out_of_range) {
      if (sn >= numFeatures())
	 throw out_of_range("Sensor index out of range");
      return *sensor[sn];
   }

private:
   // helper methods
   bool registerChipset(const sensors_chip_name& chipset);
   tempval getLimit(const sensors_chip_name& chipset, 
		    const string& key, const string& limit_suffix_str,
		    int &nr1, int &nr2);
};

#endif
